@@ -1,22 +1,13 @@
 <?php
 /**
- * An interface to a POP3 server using PHP functions.
+ * Copyright 2009-2012 Horde LLC (http://www.horde.org/)
  *
- * It is an abstraction layer allowing POP3 commands to be used based on
- * IMAP equivalents.
- *
- * This driver implements the following POP3-related RFCs:
- *   - STD 53/RFC 1939: POP3 specification
- *   - RFC 2195: CRAM-MD5 authentication
- *   - RFC 2449: POP3 extension mechanism
- *   - RFC 2595/4616: PLAIN authentication
- *   - RFC 2831: DIGEST-MD5 SASL Authentication (obsoleted by RFC 6331)
- *   - RFC 3206: AUTH/SYS response codes
- *   - RFC 1734/5034: POP3 SASL
+ * See the enclosed file COPYING for license information (LGPL). If you
+ * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
  * ---------------------------------------------------------------------------
  *
- * Originally based on the PEAR Net_POP3 package (version 1.3.6) by:
+ * Based on the PEAR Net_POP3 package (version 1.3.6) by:
  *     Richard Heyes <richard@phpguru.org>
  *     Damian Fernandez Sosa <damlists@cnba.uba.ar>
  *
@@ -50,15 +41,35 @@
  *
  * ---------------------------------------------------------------------------
  *
- * Copyright 2009-2012 Horde LLC (http://www.horde.org/)
+ * @category  Horde
+ * @copyright 2002 Richard Heyes
+ * @copyright 2009-2012 Horde LLC
+ * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
+ * @package   Imap_Client
+ */
+
+/**
+ * An interface to a POP3 server using PHP functions.
  *
- * See the enclosed file COPYING for license information (LGPL). If you
- * did not receive this file, see http://www.horde.org/licenses/lgpl21.
+ * It is an abstraction layer allowing POP3 commands to be used based on
+ * IMAP equivalents.
  *
- * @author   Michael Slusarz <slusarz@horde.org>
- * @category Horde
- * @license  http://www.horde.org/licenses/lgpl21 LGPL 2.1
- * @package  Imap_Client
+ * This driver implements the following POP3-related RFCs:
+ *   - STD 53/RFC 1939: POP3 specification
+ *   - RFC 2195: CRAM-MD5 authentication
+ *   - RFC 2449: POP3 extension mechanism
+ *   - RFC 2595/4616: PLAIN authentication
+ *   - RFC 2831: DIGEST-MD5 SASL Authentication (obsoleted by RFC 6331)
+ *   - RFC 3206: AUTH/SYS response codes
+ *   - RFC 1734/5034: POP3 SASL
+ *
+ * @author    Richard Heyes <richard@phpguru.org>
+ * @author    Michael Slusarz <slusarz@horde.org>
+ * @category  Horde
+ * @copyright 2002 Richard Heyes
+ * @copyright 2009-2012 Horde LLC
+ * @license   http://www.horde.org/licenses/lgpl21 LGPL 2.1
+ * @package   Imap_Client
  */
 class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
 {
@@ -83,21 +94,17 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
      */
     protected $_stream = null;
 
-    /**
-     */
-    protected $_utilsClass = 'Horde_Imap_Client_Utils_Pop3';
-
     /**
      */
     public function __construct(array $params = array())
     {
+        parent::__construct($params);
+
         if (empty($params['port'])) {
-            $params['port'] = (isset($params['secure']) && ($params['secure'] == 'ssl'))
+            $this->_params['port'] = (isset($this->_params['secure']) && in_array($this->_params['secure'], array('ssl', 'sslv2', 'sslv3')))
                 ? 995
                 : 110;
         }
-
-        parent::__construct($params);
     }
 
     /**
@@ -164,10 +171,11 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getNamespaces()
     {
-        $this->_exception('IMAP namespaces not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Namespaces');
     }
 
     /**
@@ -188,7 +196,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             ($this->_params['secure'] == 'tls')) {
             // Switch over to a TLS connection.
             if (!$this->queryCapability('STLS')) {
-                $this->_exception(Horde_Imap_Client_Translation::t("Could not open secure connection to the POP3 server.") . ' ' . Horde_Imap_Client_Translation::t("Server does not support secure connections."), 'LOGIN_TLSFAILURE');
+                throw new Horde_Imap_Client_Exception(
+                    Horde_Imap_Client_Translation::t("Could not open secure connection to the POP3 server.") . ' ' . Horde_Imap_Client_Translation::t("Server does not support secure connections."),
+                    Horde_Imap_Client_Exception::LOGIN_TLSFAILURE
+                );
             }
 
             $this->_sendLine('STLS');
@@ -197,7 +208,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
 
             if (!$res) {
                 $this->logout();
-                $this->_exception(Horde_Imap_Client_Translation::t("Could not open secure connection to the POP3 server."), 'LOGIN_TLSFAILURE');
+                throw new Horde_Imap_Client_Exception(
+                    Horde_Imap_Client_Translation::t("Could not open secure connection to the POP3 server."),
+                    Horde_Imap_Client_Exception::LOGIN_TLSFAILURE
+                );
             }
 
             // Expire cached CAPABILITY information
@@ -227,13 +241,16 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 return true;
             } catch (Horde_Imap_Client_Exception $e) {
                 if (!empty($this->_init['authmethod'])) {
-                    $this->_setInit('authmethod');
+                    $this->_setInit();
                     return $this->login();
                 }
             }
         }
 
-        $this->_exception(Horde_Imap_Client_Translation::t("POP3 server denied authentication."), $e->getCode() ? $e->getCode() : 'LOGIN_AUTHENTICATIONFAILED');
+        throw new Horde_Imap_Client_Exception(
+            Horde_Imap_Client_Translation::t("POP3 server denied authentication."),
+            $e->getCode() ? $e->getCode() : Horde_Imap_Client_Exception::LOGIN_AUTHENTICATIONFAILED
+        );
     }
 
     /**
@@ -253,7 +270,9 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
 
         switch ($this->_params['secure']) {
         case 'ssl':
-            $conn = 'ssl://';
+        case 'sslv2':
+        case 'sslv3':
+            $conn = $this->_params['secure'] . '://';
             $this->_isSecure = true;
             break;
 
@@ -268,10 +287,12 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
         if ($this->_stream === false) {
             $this->_stream = null;
             $this->_isSecure = false;
-            $this->_exception(array(
+            $e = new Horde_Imap_Client_Exception(
                 Horde_Imap_Client_Translation::t("Error connecting to POP3 server."),
-                sprintf("[%u] %s.", $error_number, $error_string)
-            ), 'SERVER_CONNECT');
+                Horde_Imap_Client_Exception::SERVER_CONNECT
+            );
+            $e->details = sprintf("[%u] %s.", $error_number, $error_string);
+            throw $e;
         }
 
         stream_set_timeout($this->_stream, $this->_params['timeout']);
@@ -300,8 +321,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             // RFC 5034: CRAM-MD5
             // CRAM-SHA1 & CRAM-SHA256 supported by Courier SASL library
             $challenge = $this->_sendLine('AUTH ' . $method);
-            $response = base64_encode($this->_params['username'] . ' ' . hash_hmac(strtolower($method, 5), $this->getParam('password'), base64_decode(substr($challenge['line'], 2)), true));
-            $this->_sendLine($response, array('debug' => '[' . $method . ' Response]'));
+            $response = base64_encode($this->_params['username'] . ' ' . hash_hmac(strtolower(substr($method, 5)), base64_decode(substr($challenge['line'], 2)), $this->getParam('password'), true));
+            $this->_sendLine($response, array(
+                'debug' => '[' . $method . ' Response]'
+            ));
             break;
 
         case 'DIGEST-MD5':
@@ -318,7 +341,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 'debug' => '[DIGEST-MD5 Response]'
             ));
             if (stripos(base64_decode(substr($sresponse['line'], 2)), 'rspauth=') === false) {
-                $this->_exception(Horde_Imap_Client_Translation::t("Unexpected response from server when authenticating."), 'SERVER_CONNECT');
+                throw new Horde_Imap_Client_Exception(
+                    Horde_Imap_Client_Translation::t("Unexpected response from server when authenticating."),
+                    Horde_Imap_Client_Exception::SERVER_CONNECT
+                );
             }
 
             /* POP3 doesn't use protocol's third step. */
@@ -336,7 +362,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
 
         case 'PLAIN':
             // RFC 5034
-            $this->_sendLine('AUTH PLAIN ' . base64_encode(chr(0) . $this->_params['username'] . chr(0) . $this->getParam('password')), array(
+            $this->_sendLine('AUTH PLAIN ' . base64_encode(implode("\0", array($this->_params['username'], $this->getParam('password')))), array(
                 'debug' => sprintf('[AUTH PLAIN Command - username: %s]', $this->_params['username'])
             ));
             break;
@@ -355,7 +381,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             break;
 
         default:
-            $this->_exception(sprintf(Horde_Imap_Client_Translation::t("Unknown authentication method: %s"), $method), 'SERVER_CONNECT');
+            throw new Horde_Imap_Client_Exception(
+                sprintf(Horde_Imap_Client_Translation::t("Unknown authentication method: %s"), $method),
+                Horde_Imap_Client_Exception::SERVER_CONNECT
+            );
         }
     }
 
@@ -374,10 +403,11 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _sendID($info)
     {
-        $this->_exception('IMAP ID command not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ID command');
     }
 
     /**
@@ -392,56 +422,64 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _setLanguage($langs)
     {
-        $this->_exception('IMAP LANGUAGE extension not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('LANGUAGE extension');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getLanguage($list)
     {
-        $this->_exception('IMAP LANGUAGE extension not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('LANGUAGE extension');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _openMailbox(Horde_Imap_Client_Mailbox $mailbox, $mode)
     {
         if (strcasecmp($mailbox, 'INBOX') !== 0) {
-            $this->_exception('Mailboxes other than INBOX not supported on POP3 servers.', 'NO_SUPPORT');
+            throw new Horde_Imap_Client_Exception_NoSupportPop3('Mailboxes other than INBOX');
         }
+        $this->_changeSelected($mailbox, $mode);
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _createMailbox(Horde_Imap_Client_Mailbox $mailbox, $opts)
     {
-        $this->_exception('Creating mailboxes not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Creating mailboxes');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _deleteMailbox(Horde_Imap_Client_Mailbox $mailbox)
     {
-        $this->_exception('Deleting mailboxes not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Deleting mailboxes');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _renameMailbox(Horde_Imap_Client_Mailbox $old,
                                       Horde_Imap_Client_Mailbox $new)
     {
-        $this->_exception('Renaming mailboxes not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Renaming mailboxes');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _subscribeMailbox(Horde_Imap_Client_Mailbox $mailbox,
                                          $subscribe)
     {
-        $this->_exception('Mailboxes other than INBOX not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Mailboxes other than INBOX');
     }
 
     /**
@@ -465,20 +503,12 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     /**
      * @param integer $flags   This driver only supports the options listed
      *                         under Horde_Imap_Client::STATUS_ALL.
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _status(Horde_Imap_Client_Mailbox $mailbox, $flags)
     {
         $this->openMailbox($mailbox);
 
-        // This driver only supports the base flags given by c-client.
-        if (($flags & Horde_Imap_Client::STATUS_FIRSTUNSEEN) ||
-            ($flags & Horde_Imap_Client::STATUS_FLAGS) ||
-            ($flags & Horde_Imap_Client::STATUS_PERMFLAGS) ||
-            ($flags & Horde_Imap_Client::STATUS_HIGHESTMODSEQ) ||
-            ($flags & Horde_Imap_Client::STATUS_UIDNOTSTICKY)) {
-            $this->_exception('Improper status request on POP3 server.', 'NO_SUPPORT');
-        }
-
         $ret = array();
 
         if ($flags & Horde_Imap_Client::STATUS_MESSAGES) {
@@ -491,6 +521,8 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             $ret['recent'] = $res['msgs'];
         }
 
+        // No need for STATUS_UIDNEXT_FORCE handling since STATUS_UIDNEXT will
+        // always return a value.
         if ($flags & Horde_Imap_Client::STATUS_UIDNEXT) {
             $res = $this->_pop3Cache('stat');
             $ret['uidnext'] = $res['msgs'] + 1;
@@ -510,11 +542,12 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _append(Horde_Imap_Client_Mailbox $mailbox, $data,
                                $options)
     {
-        $this->_exception('Appending messages not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Appending messages');
     }
 
     /**
@@ -547,6 +580,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _search($query, $options)
     {
@@ -555,11 +589,11 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             : reset($options['sort']);
 
         // Only support a single query: an ALL search sorted by sequence.
-        if ((reset($options['_query']['query']) != 'ALL') ||
+        if ((strval($options['_query']['query']) != 'ALL') ||
             ($sort &&
              ((count($options['sort']) > 1) ||
               ($sort != Horde_Imap_Client::SORT_SEQUENCE)))) {
-            $this->_exception('Server search not supported on POP3 server.', 'NO_SUPPORT');
+            throw new Horde_Imap_Client_Exception_NoSupportPop3('Server search');
         }
 
         $status = $this->status($this->_selected, Horde_Imap_Client::STATUS_MESSAGES);
@@ -599,41 +633,41 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _setComparator($comparator)
     {
-        $this->_exception('Search comparators not supported on POP3 server.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Search comparators');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getComparator()
     {
-        $this->_exception('Search comparators not supported on POP3 server.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Search comparators');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _thread($options)
     {
-        $this->_exception('Server threading not supported on POP3 server.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Server threading');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
-    protected function _fetch($query, $results, $options)
+    protected function _fetch(Horde_Imap_Client_Fetch_Results $results,
+                              Horde_Imap_Client_Fetch_Query $query,
+                              $options)
     {
-        // These options are not supported by this driver.
-        if (!empty($options['changedsince']) ||
-            !empty($options['vanished'])) {
-            $this->_exception('Fetch options not supported on POP3 server.', 'NO_SUPPORT');
-        }
-
         // Grab sequence IDs - IDs will always be the message number for
         // POP3 fetch commands.
         $seq_ids = $this->_getSeqIds($options['ids']);
         if (empty($seq_ids)) {
-            return $results;
+            return;
         }
 
         $lookup = $options['ids']->sequence
@@ -649,9 +683,9 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                     if (empty($c_val['start']) && empty($c_val['length'])) {
                         $tmp2 = fopen('php://temp', 'r+');
                         stream_copy_to_stream($tmp, $tmp2, empty($c_val['length']) ? -1 : $c_val['length'], empty($c_val['start']) ? 0 : $c_val['start']);
-                        $results[$lookup[$id]]->setFullMsg($tmp2);
+                        $results->get($lookup[$id])->setFullMsg($tmp2);
                     } else {
-                        $results[$lookup[$id]]->setFullMsg($tmp);
+                        $results->get($lookup[$id])->setFullMsg($tmp);
                     }
                 }
                 break;
@@ -666,7 +700,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                             $tmp = ($key == 0)
                                 ? $this->_pop3Cache('hdr', $id)
                                 : Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'header', $key);
-                            $results[$lookup[$id]]->setHeaderText($key, $this->_processString($tmp, $c_val));
+                            $results->get($lookup[$id])->setHeaderText($key, $this->_processString($tmp, $c_val));
                         } catch (Horde_Mime_Exception $e) {}
                     }
                 }
@@ -677,7 +711,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 foreach ($c_val as $key => $val) {
                     foreach ($seq_ids as $id) {
                         try {
-                            $results[$lookup[$id]]->setBodyText($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'body', $key), $val));
+                            $results->get($lookup[$id])->setBodyText($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'body', $key), $val));
                         } catch (Horde_Mime_Exception $e) {}
                     }
                 }
@@ -688,7 +722,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 foreach ($c_val as $key => $val) {
                     foreach ($seq_ids as $id) {
                         try {
-                            $results[$lookup[$id]]->setMimeHeader($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'header', $key), $val));
+                            $results->get($lookup[$id])->setMimeHeader($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'header', $key), $val));
                         } catch (Horde_Mime_Exception $e) {}
                     }
                 }
@@ -699,7 +733,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 foreach ($c_val as $key => $val) {
                     foreach ($seq_ids as $id) {
                         try {
-                            $results[$lookup[$id]]->setBodyPart($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'body', $key), $val));
+                            $results->get($lookup[$id])->setBodyPart($key, $this->_processString(Horde_Mime_Part::getRawPartText(stream_get_contents($this->_pop3Cache('msg', $id)), 'body', $key), $val));
                         } catch (Horde_Mime_Exception $e) {}
                     }
                 }
@@ -725,7 +759,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                             }
                         }
 
-                        $results[$lookup[$id]]->setHeaders($key, $tmp);
+                        $results->get($lookup[$id])->setHeaders($key, $tmp);
                     }
                 }
                 break;
@@ -734,7 +768,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 foreach ($seq_ids as $id) {
                     if ($ptr = $this->_pop3Cache('msg', $id)) {
                         try {
-                            $results[$lookup[$id]]->setStructure(Horde_Mime_Part::parseMessage(stream_get_contents($ptr)));
+                            $results->get($lookup[$id])->setStructure(Horde_Mime_Part::parseMessage(stream_get_contents($ptr)));
                         } catch (Horde_Exception $e) {}
                     }
                 }
@@ -743,7 +777,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             case Horde_Imap_Client::FETCH_ENVELOPE:
                 foreach ($seq_ids as $id) {
                     $tmp = $this->_pop3Cache('hdrob', $id);
-                    $results[$lookup[$id]]->setEnvelope(array(
+                    $results->get($lookup[$id])->setEnvelope(array(
                         'date' => $tmp->getValue('date'),
                         'subject' => $tmp->getValue('subject'),
                         'from' => $tmp->getOb('from'),
@@ -761,20 +795,20 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             case Horde_Imap_Client::FETCH_IMAPDATE:
                 foreach ($seq_ids as $id) {
                     $tmp = $this->_pop3Cache('hdrob', $id);
-                    $results[$lookup[$id]]->setImapDate($tmp->getValue('date'));
+                    $results->get($lookup[$id])->setImapDate($tmp->getValue('date'));
                 }
                 break;
 
             case Horde_Imap_Client::FETCH_SIZE:
                 $sizelist = $this->_pop3Cache('size');
                 foreach ($seq_ids as $id) {
-                    $results[$lookup[$id]]->setSize($sizelist[$id]);
+                    $results->get($lookup[$id])->setSize($sizelist[$id]);
                 }
                 break;
 
             case Horde_Imap_Client::FETCH_SEQ:
                 foreach ($seq_ids as $id) {
-                    $results[$lookup[$id]]->setSeq($id);
+                    $results->get($lookup[$id])->setSeq($id);
                 }
                 break;
 
@@ -782,18 +816,14 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
                 $uidllist = $this->_pop3Cache('uidl');
                 foreach ($seq_ids as $id) {
                     if (isset($uidllist[$id])) {
-                        $results[$lookup[$id]]->setUid($uidllist[$id]);
+                        $results->get($lookup[$id])->setUid($uidllist[$id]);
                     }
                 }
                 break;
             }
         }
 
-        $this->_updateCache($results, array(
-            'seq' => $options['ids']->sequence
-        ));
-
-        return $results;
+        $this->_updateCache($results);
     }
 
     /**
@@ -889,6 +919,14 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
         return $str;
     }
 
+    /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
+     */
+    protected function _vanished($modseq, Horde_Imap_Client_Ids $ids)
+    {
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('QRESYNC commands');
+    }
+
     /**
      * @param array $options  Additional options. This driver does not support
      *                        'unchangedsince'.
@@ -931,81 +969,99 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _copy(Horde_Imap_Client_Mailbox $dest, $options)
     {
-        $this->_exception('Copying messages not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Copying messages');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _setQuota(Horde_Imap_Client_Mailbox $root, $options)
     {
-        $this->_exception('IMAP quotas not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Quotas');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getQuota(Horde_Imap_Client_Mailbox $root)
     {
-        $this->_exception('IMAP quotas not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Quotas');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getQuotaRoot(Horde_Imap_Client_Mailbox $mailbox)
     {
-        $this->_exception('IMAP quotas not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Quotas');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _setACL(Horde_Imap_Client_Mailbox $mailbox, $identifier,
                                $options)
     {
-        $this->_exception('IMAP ACLs not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ACLs');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
+     */
+    protected function _deleteACL(Horde_Imap_Client_Mailbox $mailbox, $identifier)
+    {
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ACLs');
+    }
+
+    /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getACL(Horde_Imap_Client_Mailbox $mailbox)
     {
-        $this->_exception('IMAP ACLs not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ACLs');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _listACLRights(Horde_Imap_Client_Mailbox $mailbox,
                                       $identifier)
     {
-        $this->_exception('IMAP ACLs not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ACLs');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getMyACLRights(Horde_Imap_Client_Mailbox $mailbox)
     {
-        $this->_exception('IMAP ACLs not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('ACLs');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _getMetadata(Horde_Imap_Client_Mailbox $mailbox,
                                     $entries, $options)
     {
-        $this->_exception('IMAP metadata not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Metadata');
     }
 
     /**
+     * @throws Horde_Imap_Client_Exception_NoSupportPop3
      */
     protected function _setMetadata(Horde_Imap_Client_Mailbox $mailbox, $data)
     {
-        $this->_exception('IMAP metadata not supported on POP3 servers.', 'NO_SUPPORT');
+        throw new Horde_Imap_Client_Exception_NoSupportPop3('Metadata');
     }
 
     /**
      */
-    protected function _getSearchCache($type, $mailbox, $options)
+    protected function _getSearchCache($type, $options)
     {
         /* POP3 does not support search caching. */
         return null;
@@ -1025,7 +1081,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
      */
     protected function _sendLine($query, $options = array())
     {
-        $this->writeDebug((empty($options['debug']) ? $query : $options['debug']) . "\n", Horde_Imap_Client::DEBUG_CLIENT);
+        $this->_debug->client(empty($options['debug']) ? $query : $options['debug']);
 
         fwrite($this->_stream, $query . "\r\n");
 
@@ -1047,7 +1103,10 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
 
         if (feof($this->_stream)) {
             $this->logout();
-            $this->_exception(Horde_Imap_Client_Translation::t("POP3 Server closed the connection unexpectedly."), 'DISCONNECT');
+            throw new Horde_Imap_Client_Exception(
+                Horde_Imap_Client_Translation::t("POP3 Server closed the connection unexpectedly."),
+                Horde_Imap_Client_Exception::DISCONNECT
+            );
         }
 
         $read = rtrim(fgets($this->_stream));
@@ -1055,7 +1114,7 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
             return;
         }
 
-        $this->writeDebug($read . "\n", Horde_Imap_Client::DEBUG_SERVER);
+        $this->_debug->server($read);
 
         $orig_read = $read;
         $read = explode(' ', $read, 2);
@@ -1064,49 +1123,60 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
         case '+OK':
             $ob['response'] = 'OK';
             if (isset($read[1])) {
-                $response = $this->_parseResponseText($read[1]);
-                $ob['line'] = $response->text;
+                if ($this->queryCapability('RESP-CODES')) {
+                    $response = $this->_parseResponseCode($read[1]);
+                    $ob['line'] = $response->text;
+                } else {
+                    $ob['line'] = $read[1];
+                }
             }
             break;
 
         case '-ERR':
             $errcode = 0;
             if (isset($read[1])) {
-                $response = $this->_parseResponseText($read[1]);
-                $errtext = $response->text;
-                if (isset($response->code)) {
-                    switch ($response->code) {
-                    // RFC 2449 [8.1.1]
-                    case 'IN-USE':
-                    // RFC 2449 [8.1.2]
-                    case 'LOGIN-DELAY':
-                        $errcode = 'LOGIN_UNAVAILABLE';
-                        break;
+                if ($this->queryCapability('RESP-CODES')) {
+                    $response = $this->_parseResponseCode($read[1]);
+                    $errtext = $response->text;
 
-                    // RFC 3206 [4]
-                    case 'SYS/TEMP':
-                        $errcode = 'POP3_TEMP_ERROR';
-                        break;
+                    if (isset($response->code)) {
+                        switch ($response->code) {
+                        // RFC 2449 [8.1.1]
+                        case 'IN-USE':
+                        // RFC 2449 [8.1.2]
+                        case 'LOGIN-DELAY':
+                            $errcode = Horde_Imap_Client_Exception::LOGIN_UNAVAILABLE;
+                            break;
 
-                    // RFC 3206 [4]
-                    case 'SYS/PERM':
-                        $errcode = 'POP3_PERM_ERROR';
-                        break;
+                        // RFC 3206 [4]
+                        case 'SYS/TEMP':
+                            $errcode = Horde_Imap_Client_Exception::POP3_TEMP_ERROR;
+                            break;
 
-                    // RFC 3206 [5]
-                    case 'AUTH':
-                        $errcode = 'LOGIN_AUTHENTICATIONFAILED';
-                        break;
+                        // RFC 3206 [4]
+                        case 'SYS/PERM':
+                            $errcode = Horde_Imap_Client_Exception::POP3_PERM_ERROR;
+                            break;
+
+                        // RFC 3206 [5]
+                        case 'AUTH':
+                            $errcode = Horde_Imap_Client_Exception::LOGIN_AUTHENTICATIONFAILED;
+                            break;
+                        }
                     }
+                } else {
+                    $errtext = $read[1];
                 }
             } else {
                 $errtext = '[No error message provided by server]';
             }
 
-            $this->_exception(array(
+            throw new Horde_Imap_Client_Exception_ServerResponse(
                 Horde_Imap_Client_Translation::t("POP3 error reported by server."),
+                $errcode,
+                '-ERR',
                 $errtext
-            ), $errcode);
+            );
 
         case '.':
             $ob['response'] = 'END';
@@ -1177,4 +1247,36 @@ class Horde_Imap_Client_Socket_Pop3 extends Horde_Imap_Client_Base
         return array_keys(array_intersect($this->_pop3Cache('uidl'), $ids->ids));
     }
 
+    /**
+     * Parses response text for response codes (RFC 2449 [8]).
+     *
+     * @param string $text  The response text.
+     *
+     * @return object  An object with the following properties:
+     *   - code: (string) The response code, if it exists.
+     *   - data: (string) The response code data, if it exists.
+     *   - text: (string) The human-readable response text.
+     */
+    protected function _parseResponseCode($text)
+    {
+        $ret = new stdClass;
+
+        $text = trim($text);
+        if ($text[0] == '[') {
+            $pos = strpos($text, ' ', 2);
+            $end_pos = strpos($text, ']', 2);
+            if ($pos > $end_pos) {
+                $ret->code = strtoupper(substr($text, 1, $end_pos - 1));
+            } else {
+                $ret->code = strtoupper(substr($text, 1, $pos - 1));
+                $ret->data = substr($text, $pos + 1, $end_pos - $pos - 1);
+            }
+            $ret->text = trim(substr($text, $end_pos + 1));
+        } else {
+            $ret->text = $text;
+        }
+
+        return $ret;
+    }
+
 }
