@@ -1,29 +1,23 @@
 <?php
 /**
- * Tests for the Search Query object.
- *
- * PHP version 5
- *
- * @category Horde
- * @package  Imap_Client
- * @author   Michael Slusarz <slusarz@horde.org>
- * @license  http://www.horde.org/licenses/lgpl21 LGPL 2.1
- * @link     http://pear.horde.org/index.php?package=Imap_Client
- */
-
-/**
- * Tests for the Search Query object.
- *
  * Copyright 2011-2012 Horde LLC (http://www.horde.org/)
  *
  * See the enclosed file COPYING for license information (LGPL). If you
  * did not receive this file, see http://www.horde.org/licenses/lgpl21.
  *
  * @category Horde
- * @package  Imap_Client
- * @author   Michael Slusarz <slusarz@horde.org>
  * @license  http://www.horde.org/licenses/lgpl21 LGPL 2.1
- * @link     http://pear.horde.org/index.php?package=Imap_Client
+ * @package  Imap_Client
+ */
+
+/**
+ * Tests for the Search Query object.
+ *
+ * @author   Michael Slusarz <slusarz@horde.org>
+ * @category Horde
+ * @ignore
+ * @license  http://www.horde.org/licenses/lgpl21 LGPL 2.1
+ * @package  Imap_Client
  */
 class Horde_Imap_Client_SearchTest extends PHPUnit_Framework_TestCase
 {
@@ -47,4 +41,28 @@ class Horde_Imap_Client_SearchTest extends PHPUnit_Framework_TestCase
         );
     }
 
+    public function testOrQueriesWithABaseQuery()
+    {
+        $or_ob = new Horde_Imap_Client_Search_Query();
+
+        $ob = new Horde_Imap_Client_Search_Query();
+        $ob->flag('\\deleted', false);
+        $ob->headerText('from', 'ABC');
+        $or_ob->orSearch($ob);
+
+        $ob = new Horde_Imap_Client_Search_Query();
+        $ob->flag('\\deleted', true);
+        $ob->headerText('from', 'DEF');
+        $or_ob->orSearch($ob);
+
+        $base_ob = new Horde_Imap_Client_Search_Query();
+        $base_ob->flag('\\seen', false);
+        $base_ob->andSearch($or_ob);
+
+        $this->assertEquals(
+            'UNSEEN OR (DELETED FROM DEF) (UNDELETED FROM ABC)',
+             strval($base_ob)
+         );
+    }
+
 }
